from typing import Callable, Sequence, Dict, Union

import numpy as np

from ..base_transform import EEGTransform


class Concatenate(EEGTransform):
    r'''
    Merge the calculation results of multiple transforms, which are used when feature fusion is required.

    .. code-block:: python

        transform = Concatenate([
            BandDifferentialEntropy(),
            BandMeanAbsoluteDeviation()
        ])
        transform(eeg=np.random.randn(32, 128))['eeg'].shape
        >>> (32, 8)

    Args:
        transforms (list, tuple): a sequence of transforms
        apply_to_baseline: (bool): Whether to act on the baseline signal at the same time, if the baseline is passed in when calling. (defualt: :obj:`False`)
    
    .. automethod:: __call__
    '''
    def __init__(self,
                 transforms: Sequence[Callable],
                 apply_to_baseline: bool = False):
        super(Concatenate, self).__init__(apply_to_baseline=apply_to_baseline)
        self.transforms = transforms

    def __call__(self,
                 *args,
                 eeg: np.ndarray,
                 baseline: Union[np.ndarray, None] = None,
                 **kwargs) -> Dict[str, np.ndarray]:
        r'''
        Args:
            eeg (np.ndarray): The input EEG signals in shape of [number of electrodes, number of data points].
            baseline (np.ndarray, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.
            
        Returns:
            np.ndarray: The combined results of multiple transforms.
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    def apply(self, eeg: np.ndarray, **kwargs) -> np.ndarray:
        out = []
        for t in self.transforms:
            out.append(t.apply(eeg, **kwargs))
        return np.concatenate(out, axis=-1)

    def __repr__(self) -> str:
        format_string = self.__class__.__name__ + '('
        for i, t in enumerate(self.transforms):
            if i:
                format_string += ','
            format_string += '\n'
            format_string += f'    {t}'
        format_string += '\n)'
        return format_string


class ChunkConcatenate(EEGTransform):
    r'''
    Divide the input EEG signal into multiple chunks according to chunk_size and overlap, and then apply transforms to each chunk, and combine the calculation results of all transforms on all chunks. It is used when feature fusion is required.

    .. code-block:: python

        transform = ChunkConcatenate([
            BandDifferentialEntropy(),
            BandMeanAbsoluteDeviation()
        ],
        chunk_size=250,
        overlap=0)
        transform(eeg=np.random.randn(64, 1000))['eeg'].shape
        >>> (64, 32)

    TorchEEG allows feature fusion at multiple scales:

    .. code-block:: python

        transform = Concatenate([
            ChunkConcatenate([
                BandDifferentialEntropy()
            ],
            chunk_size=250,
            overlap=0),  # 4 chunk * 4-dim feature
            ChunkConcatenate([
                BandDifferentialEntropy()
            ],
            chunk_size=500,
            overlap=0),  # 2 chunk * 4-dim feature
            BandDifferentialEntropy()  # 1 chunk * 4-dim feature
        ])
        transform(eeg=np.random.randn(64, 1000))['eeg'].shape
        >>> (64, 28) # 4 * 4 + 2 * 4 + 1 * 4

    Args:
        transforms (list, tuple): a sequence of transforms
        apply_to_baseline: (bool): Whether to act on the baseline signal at the same time, if the baseline is passed in when calling. (defualt: :obj:`False`)
    
    .. automethod:: __call__
    '''
    def __init__(self,
                 transforms: Sequence[Callable],
                 chunk_size: int = 250,
                 overlap: int = 0,
                 apply_to_baseline: bool = False):
        super(ChunkConcatenate,
              self).__init__(apply_to_baseline=apply_to_baseline)
        self.transforms = transforms
        self.chunk_size = chunk_size
        self.overlap = overlap

    def __call__(self,
                 *args,
                 eeg: np.ndarray,
                 baseline: Union[np.ndarray, None] = None,
                 **kwargs) -> Dict[str, np.ndarray]:
        r'''
        Args:
            eeg (np.ndarray): The input EEG signals in shape of [number of electrodes, number of data points].
            baseline (np.ndarray, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.
            
        Returns:
            np.ndarray: The combined results of multiple transforms.
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    def apply(self, eeg: np.ndarray, **kwargs) -> np.ndarray:
        start_at = 0
        end_at = start_at + self.chunk_size
        step = self.chunk_size - self.overlap

        out = []

        while end_at <= eeg.shape[1]:

            for t in self.transforms:
                out.append(t.apply(eeg[:, start_at:end_at], **kwargs))

            start_at = start_at + step
            end_at = start_at + self.chunk_size

        return np.concatenate(out, axis=-1)

    def __repr__(self) -> str:
        format_string = self.__class__.__name__ + '(['
        for i, t in enumerate(self.transforms):
            if i:
                format_string += ','
            format_string += '\n'
            format_string += f'    {t}'
        format_string += '\n],'
        format_string += f'\nchunk_size={self.chunk_size}, '
        format_string += f'\noverlap={self.overlap})'
        return format_string



import contextlib
from typing import Dict

import torch
import torch.nn as nn
import torch.nn.functional as F


@contextlib.contextmanager
def _disable_tracking_bn_stats(model: nn.Module):
    def switch_attr(m):
        if hasattr(m, 'track_running_stats'):
            m.track_running_stats ^= True

    model.apply(switch_attr)
    yield
    model.apply(switch_attr)


def _l2_normalize(d: torch.Tensor, eps=1e-8):
    d_reshaped = d.view(d.shape[0], -1, *(1 for _ in range(d.dim() - 2)))
    d /= torch.norm(d_reshaped, dim=1, keepdim=True) + eps
    return d


class VirtualAdversarialTrainingLoss(nn.Module):
    r"""
    Virtual adversarial training loss is a regularization method based on virtual adversarial loss defined as the robustness of the conditional label distribution around each input data point against local perturbation. The virtual adversarial training loss smooth the model are only virtually adversarial and does not require label information and is hence applicable to semi-supervised learning.

    - Paper: Miyato T, Maeda S, Koyama M, et al. Virtual adversarial training: a regularization method for supervised and semi-supervised learning[J]. IEEE transactions on pattern analysis and machine intelligence, 2018, 41(8): 1979-1993.
    - URL: https://ieeexplore.ieee.org/abstract/document/8417973
    - Related Project: https://github.com/lyakaap/VAT-pytorch

    .. code-block:: python

        inputs = torch.randn(3, 5, requires_grad=True)
        model = nn.Linear(5, 5)
        loss = VirtualAdversarialTrainingLoss()
        output = loss(model, inputs)

    Args:
        xi (float): The hyperparameter xi in the focal loss. (defualt: :obj:`10.0`)
        eps (float): The hyperparameter eps in the focal loss. (defualt: :obj:`1.0`)
        iterations (int): iteration times of computing adversarial noise. (defualt: :obj:`1`)
    """
    def __init__(self, xi: float = 10.0, eps: float = 1.0, iterations: int = 1):
        super(VirtualAdversarialTrainingLoss, self).__init__()
        self.xi = xi
        self.eps = eps
        self.iterations = iterations

    def forward(self,
                model: nn.Module,
                x: torch.Tensor) -> torch.Tensor:
        r'''
        Args:
            model (nn.Module): For the model used for classification, the input of the forward function of the model should be a torch.Tensor, and the output should be a torch.Tensor corresponding to the predictions.
            x (torch.Tensor): The input data for the model.

        Returns:
            torch.Tensor: The loss value.
        '''
        with torch.no_grad():
            pred = F.softmax(model(x), dim=1)

        d = torch.rand(x.shape).sub(0.5).to(x.device)
        d = _l2_normalize(d)

        with _disable_tracking_bn_stats(model):

            for _ in range(self.iterations):
                d.requires_grad_()
                pred_hat = model(x + self.xi * d)
                logp_hat = F.log_softmax(pred_hat, dim=1)
                adv_distance = F.kl_div(logp_hat, pred, reduction='batchmean')
                adv_distance.backward()
                d = _l2_normalize(d.grad)
                model.zero_grad()

            r_adv = d * self.eps
            pred_hat = model(x + r_adv)
            logp_hat = F.log_softmax(pred_hat, dim=1)
            lds = F.kl_div(logp_hat, pred, reduction='batchmean')

        return lds

    @property
    def repr_body(self) -> Dict:
        return {'xi': self.xi, 'eps': self.eps, 'iterations': self.iterations}

    def __repr__(self) -> str:
        # init info
        format_string = self.__class__.__name__ + '('
        for i, (k, v) in enumerate(self.repr_body.items()):
            # line end
            if i:
                format_string += ', '
            # str param
            if isinstance(v, str):
                format_string += f"{k}='{v}'"
            else:
                format_string += f"{k}={v}"
        format_string += ')'
        return format_string



from typing import List, Union, Dict

from ..base_transform import LabelTransform


class Binary(LabelTransform):
    r'''
    Binarize the label according to a certain threshold. Labels larger than the threshold are set to 1, and labels smaller than the threshold are set to 0.
    
    .. code-block:: python

        transform = Binary(threshold=5.0)
        transform(y=4.5)['y']
        >>> 0

    :obj:`Binary` allows simultaneous binarization using the same threshold for multiple labels.

    .. code-block:: python

        transform = Binary(threshold=5.0)
        transform(y=[4.5, 5.5])['y']
        >>> [0, 1]

    Args:
        threshold (float): Threshold used during binarization.

    .. automethod:: __call__
    '''
    def __init__(self, threshold: float):
        super(Binary, self).__init__()
        self.threshold = threshold

    def __call__(self, *args, y: Union[int, float, List],
                 **kwargs) -> Union[int, List]:
        r'''
        Args:
            label (int, float, or list): The input label or list of labels.
            
        Returns:
            int, float, or list: The output label or list of labels after binarization.
        '''
        return super().__call__(*args, y=y, **kwargs)

    def apply(self, y: Union[int, float, List], **kwargs) -> Union[int, List]:
        assert isinstance(
            y, (int, float, list)
        ), f'The transform Binary only accepts label list or item (int or float) as input, but obtain {type(y)} as input.'
        if isinstance(y, list):
            return [int(l >= self.threshold) for l in y]
        return int(y >= self.threshold)

    @property
    def repr_body(self) -> Dict:
        return dict(super().repr_body, **{
            'threshold': self.threshold
        })

class BinariesToCategory(LabelTransform):
    r'''
    Convert multiple binary labels into one multiclass label. Multiclass labels represent permutations of binary labels. Commonly used to combine two binary classification tasks into a single quad classification task.
    
    .. code-block:: python

        transform = BinariesToCategory()
        transform(y=[0, 0])['y']
        >>> 0
        transform(y=[0, 1])['y']
        >>> 1
        transform(y=[1, 0])['y']
        >>> 2
        transform(y=[1, 1])['y']
        >>> 3
    
    .. automethod:: __call__
    '''
    def __call__(self, *args, y: List, **kwargs) -> int:
        r'''
        Args:
            y (list): list of binary labels.
            
        Returns:
            int: The converted multiclass label.
        '''
        return super().__call__(*args, y=y, **kwargs)

    def apply(self, y: List, **kwargs) -> int:
        r'''
        Args:
            y (list): list of binary labels.
            
        Returns:
            int: The converted multiclass label.
        '''
        assert isinstance(
            y, list
        ), f'The transform BinariesToCategory only accepts label list as input, but obtain {type(y)} as input.'
        return sum([v * 2**i for i, v in enumerate(reversed(y))])



from .deap import deap_constructor
from .dreamer import dreamer_constructor
from .seed import seed_constructor
from .amigos import amigos_constructor
from .mahnob import mahnob_constructor
from .bci2022 import bci2022_constructor


import os
from functools import partial
from multiprocessing import Manager, Pool, Process, Queue, set_start_method
from typing import Callable, List, Union

import scipy.io as scio
from torcheeg.io import EEGSignalIO, MetaInfoIO
from tqdm import tqdm

MAX_QUEUE_SIZE = 1024


def transform_producer(file_name: str, root_path: str, chunk_size: int,
                       overlap: int, channel_num: int,
                       transform: Union[List[Callable], Callable, None],
                       write_info_fn: Callable, queue: Queue):
    subject = int(
        os.path.basename(file_name).split('.')[0].split('_')[0])  # subject (15)
    date = int(
        os.path.basename(file_name).split('.')[0].split('_')[1])  # period (3)

    samples = scio.loadmat(os.path.join(root_path, file_name),
                           verify_compressed_data_integrity=False
                           )  # trial (15), channel(62), timestep(n*200)
    # label file
    labels = scio.loadmat(os.path.join(root_path, 'label.mat'),
                          verify_compressed_data_integrity=False)['label'][0]

    trial_ids = [key for key in samples.keys() if 'eeg' in key]

    # calculate moving step
    step = chunk_size - overlap

    write_pointer = 0
    # loop for each trial
    for trial_id in trial_ids:
        # extract baseline signals
        trial_samples = samples[trial_id]  # channel(62), timestep(n*200)

        # record the common meta info
        trial_meta_info = {
            'subject_id': subject,
            'trial_id': trial_id,
            'emotion': int(labels[int(trial_id.split('_')[-1][3:]) - 1]),
            'date': date
        }

        # extract experimental signals
        start_at = 0
        end_at = chunk_size

        while end_at <= trial_samples.shape[1]:
            clip_sample = trial_samples[:channel_num, start_at:end_at]

            t_eeg = clip_sample
            if not transform is None:
                t_eeg = transform(eeg=clip_sample)['eeg']

            clip_id = f'{file_name}_{write_pointer}'

            queue.put({'eeg': t_eeg, 'key': clip_id})
            write_pointer += 1

            # record meta info for each signal
            record_info = {
                'start_at': start_at,
                'end_at': end_at,
                'clip_id': clip_id
            }
            record_info.update(trial_meta_info)
            write_info_fn(record_info)

            start_at = start_at + step
            end_at = start_at + chunk_size


def io_consumer(write_eeg_fn, queue):
    while True:
        item = queue.get()
        if not item is None:
            eeg = item['eeg']
            key = item['key']
            write_eeg_fn(eeg, key)
        else:
            break


class SingleProcessingQueue:
    def __init__(self, write_eeg_fn):
        self.write_eeg_fn = write_eeg_fn

    def put(self, item):
        eeg = item['eeg']
        key = item['key']
        self.write_eeg_fn(eeg, key)


def seed_constructor(root_path: str = './Preprocessed_EEG',
                     chunk_size: int = 200,
                     overlap: int = 0,
                     channel_num: int = 62,
                     transform: Union[None, Callable] = None,
                     io_path: str = './io/seed',
                     num_worker: int = 0,
                     verbose: bool = True,
                     cache_size: int = 64 * 1024 * 1024 * 1024) -> None:
    # init IO
    meta_info_io_path = os.path.join(io_path, 'info.csv')
    eeg_signal_io_path = os.path.join(io_path, 'eeg')

    if os.path.exists(io_path) and not os.path.getsize(meta_info_io_path) == 0:
        print(
            f'The target folder already exists, if you need to regenerate the database IO, please delete the path {io_path}.'
        )
        return

    meta_info_io_path = os.path.join(io_path, 'info.csv')
    eeg_signal_io_path = os.path.join(io_path, 'eeg')

    info_io = MetaInfoIO(meta_info_io_path)
    eeg_io = EEGSignalIO(eeg_signal_io_path, cache_size=cache_size)

    # loop to access the dataset files
    file_list = os.listdir(root_path)
    skip_set = ['label.mat', 'readme.txt']
    file_list = [f for f in file_list if f not in skip_set]

    if verbose:
        # show process bar
        pbar = tqdm(total=len(file_list))
        pbar.set_description("[SEED]")

    if num_worker > 1:
        manager = Manager()
        queue = manager.Queue(maxsize=MAX_QUEUE_SIZE)
        io_consumer_process = Process(target=io_consumer,
                                      args=(eeg_io.write_eeg, queue),
                                      daemon=True)
        io_consumer_process.start()

        partial_mp_fn = partial(transform_producer,
                                root_path=root_path,
                                chunk_size=chunk_size,
                                overlap=overlap,
                                channel_num=channel_num,
                                transform=transform,
                                write_info_fn=info_io.write_info,
                                queue=queue)

        for _ in Pool(num_worker).imap(partial_mp_fn, file_list):
            if verbose:
                pbar.update(1)

        queue.put(None)

        io_consumer_process.join()
        io_consumer_process.close()

    else:
        for file_name in file_list:
            transform_producer(file_name=file_name,
                               root_path=root_path,
                               chunk_size=chunk_size,
                               overlap=overlap,
                               channel_num=channel_num,
                               transform=transform,
                               write_info_fn=info_io.write_info,
                               queue=SingleProcessingQueue(eeg_io.write_eeg))
            if verbose:
                pbar.update(1)

    if verbose:
        pbar.close()
        print('Please wait for the writing process to complete...')



import os
import re
from functools import partial
from multiprocessing import Manager, Pool, Process, Queue, set_start_method
from typing import Callable, List, Union

import scipy.io as scio
from torcheeg.io import EEGSignalIO, MetaInfoIO
from tqdm import tqdm

MAX_QUEUE_SIZE = 1024


def transform_producer(file_name: str, root_path: str, chunk_size: int,
                       overlap: int, channel_num: int, trial_num: int,
                       skipped_subjects: List, baseline_num: int,
                       baseline_chunk_size: int,
                       transform: Union[List[Callable], Callable, None],
                       write_info_fn: Callable, queue: Queue):
    subject = int(re.findall(r'Data_Preprocessed_P(\d*).mat',
                             file_name)[0])  # subject (40)

    if subject in skipped_subjects:
        return

    data = scio.loadmat(os.path.join(root_path, file_name),
                        verify_compressed_data_integrity=False)
    samples = data['joined_data'][
        0]  # trial (20), timestep(n*128), channel(17) (14 channels are EEGs)
    # label file
    labels = data['labels_selfassessment'][
        0]  # trial (20), label of different dimensions ((1, 12))

    # calculate moving step
    step = chunk_size - overlap

    write_pointer = 0

    max_len = len(samples)
    if not (trial_num <= 0):
        max_len = min(len(samples), trial_num)

    # loop for each trial
    for trial_id in range(max_len):
        # extract baseline signals
        trial_samples = samples[trial_id]  # timestep(n*128), channel(17)

        # record the common meta info
        trial_meta_info = {'subject_id': subject, 'trial_id': trial_id}
        trial_rating = labels[trial_id][0]  # label of different dimensions (12)

        # missing values
        if (not sum(trial_samples.shape)) or (not sum(trial_rating.shape)):
            # 3 of the participants (08,24,28<->32) of the previous experiment did not watch a set of 4 long affective
            if sum(trial_samples.shape) != sum(trial_rating.shape):
                print(
                    f'[WARNING] Find EEG signals without labels, or labels without EEG signals. Please check the {trial_id + 1}-th experiment of the {subject}-th subject in the file {file_name}. TorchEEG currently skipped the mismatched data.'
                )
            continue

        for label_idx, label_name in enumerate([
                'arousal', 'valence', 'dominance', 'liking', 'familiarity',
                'neutral', 'disgust', 'happiness', 'surprise', 'anger', 'fear',
                'sadness'
        ]):
            trial_meta_info[label_name] = trial_rating[label_idx]

        # extract baseline signals
        trial_baseline_sample = trial_samples[:baseline_chunk_size *
                                              baseline_num, :
                                              channel_num]  # timestep(5*128), channel(14)
        trial_baseline_sample = trial_baseline_sample.reshape(
            baseline_num, baseline_chunk_size,
            channel_num).mean(axis=0).swapaxes(1,
                                               0)  # channel(14), timestep(128)

        # extract experimental signals
        start_at = baseline_chunk_size * baseline_num
        end_at = start_at + chunk_size

        while end_at <= trial_samples.shape[0]:
            clip_sample = trial_samples[start_at:end_at, :channel_num].swapaxes(
                1, 0)

            t_eeg = clip_sample
            t_baseline = trial_baseline_sample
            if not transform is None:
                t = transform(eeg=clip_sample, baseline=trial_baseline_sample)
                t_eeg = t['eeg']
                t_baseline = t['baseline']

            # put baseline signal into IO
            if not 'baseline_id' in trial_meta_info:
                trial_base_id = f'{file_name}_{write_pointer}'
                queue.put({'eeg': t_baseline, 'key': trial_base_id})
                write_pointer += 1
                trial_meta_info['baseline_id'] = trial_base_id

            clip_id = f'{file_name}_{write_pointer}'
            queue.put({'eeg': t_eeg, 'key': clip_id})
            write_pointer += 1

            # record meta info for each signal
            record_info = {
                'start_at': start_at,
                'end_at': end_at,
                'clip_id': clip_id
            }
            record_info.update(trial_meta_info)
            write_info_fn(record_info)

            start_at = start_at + step
            end_at = start_at + chunk_size


def io_consumer(write_eeg_fn, queue):
    while True:
        item = queue.get()
        if not item is None:
            eeg = item['eeg']
            key = item['key']
            write_eeg_fn(eeg, key)
        else:
            break


class SingleProcessingQueue:
    def __init__(self, write_eeg_fn):
        self.write_eeg_fn = write_eeg_fn

    def put(self, item):
        eeg = item['eeg']
        key = item['key']
        self.write_eeg_fn(eeg, key)


def amigos_constructor(root_path: str = './data_preprocessed',
                       chunk_size: int = 128,
                       overlap: int = 0,
                       channel_num: int = 14,
                       trial_num: int = 16,
                       skipped_subjects: List[int] = [
                           9, 12, 21, 22, 23, 24, 33
                       ],
                       baseline_num: int = 5,
                       baseline_chunk_size: int = 128,
                       transform: Union[None, Callable] = None,
                       io_path: str = './io/amigos',
                       num_worker: int = 0,
                       verbose: bool = True,
                       cache_size: int = 64 * 1024 * 1024 * 1024) -> None:
    # init IO
    meta_info_io_path = os.path.join(io_path, 'info.csv')
    eeg_signal_io_path = os.path.join(io_path, 'eeg')

    if os.path.exists(io_path) and not os.path.getsize(meta_info_io_path) == 0:
        print(
            f'The target folder already exists, if you need to regenerate the database IO, please delete the path {io_path}.'
        )
        return

    os.makedirs(io_path, exist_ok=True)

    info_io = MetaInfoIO(meta_info_io_path)
    eeg_io = EEGSignalIO(eeg_signal_io_path, cache_size=cache_size)

    # loop to access the dataset files
    file_list = os.listdir(root_path)

    if verbose:
        # show process bar
        pbar = tqdm(total=len(file_list))
        pbar.set_description("[AMIGOS]")

    if num_worker > 1:
        manager = Manager()
        queue = manager.Queue(maxsize=MAX_QUEUE_SIZE)

        partial_mp_fn = partial(transform_producer,
                                root_path=root_path,
                                chunk_size=chunk_size,
                                overlap=overlap,
                                channel_num=channel_num,
                                trial_num=trial_num,
                                skipped_subjects=skipped_subjects,
                                baseline_num=baseline_num,
                                baseline_chunk_size=baseline_chunk_size,
                                transform=transform,
                                write_info_fn=info_io.write_info,
                                queue=queue)

        io_consumer_process = Process(target=io_consumer,
                                      args=(eeg_io.write_eeg, queue),
                                      daemon=True)
        io_consumer_process.start()

        for _ in Pool(num_worker).imap(partial_mp_fn, file_list):
            if verbose:
                pbar.update(1)

        queue.put(None)

        io_consumer_process.join()
        io_consumer_process.close()

    else:
        for file_name in file_list:
            transform_producer(file_name=file_name,
                               root_path=root_path,
                               chunk_size=chunk_size,
                               overlap=overlap,
                               channel_num=channel_num,
                               trial_num=trial_num,
                               skipped_subjects=skipped_subjects,
                               baseline_num=baseline_num,
                               baseline_chunk_size=baseline_chunk_size,
                               transform=transform,
                               write_info_fn=info_io.write_info,
                               queue=SingleProcessingQueue(eeg_io.write_eeg))
            if verbose:
                pbar.update(1)

    if verbose:
        pbar.close()
        print('Please wait for the writing process to complete...')



import unittest

import torch
import torch.nn as nn

from torcheeg.losses import FocalLoss, VirtualAdversarialTrainingLoss


class TestLosses(unittest.TestCase):
    def test_focal_loss(self):
        inputs = torch.randn(3, 5, requires_grad=True)
        targets = torch.empty(3, dtype=torch.long).random_(5)
        loss = FocalLoss(num_classes=5)
        output = loss(inputs, targets)
        output.backward()
        self.assertEqual(tuple(output.shape), ())

    def test_virtual_adversarial_training_loss(self):
        inputs = torch.randn(3, 5, requires_grad=True)
        model = nn.Linear(5, 5)
        loss = VirtualAdversarialTrainingLoss()
        output = loss(model, inputs)
        output.backward()
        self.assertEqual(tuple(output.shape), ())


if __name__ == '__main__':
    unittest.main()



import os
from typing import Callable, List, Union
from functools import partial
import scipy.io as scio
from multiprocessing import Manager, Pool, Process, Queue, set_start_method
from torcheeg.io import EEGSignalIO, MetaInfoIO
from tqdm import tqdm

MAX_QUEUE_SIZE = 1024


def transform_producer(subject: int, trial_len: int, mat_data: any,
                       chunk_size: int, overlap: int, channel_num: int,
                       baseline_num: int, baseline_chunk_size: int,
                       transform: Union[List[Callable], Callable, None],
                       write_info_fn: Callable, queue: Queue):
    # calculate moving step
    step = chunk_size - overlap

    write_pointer = 0
    # loop for each trial
    for trial_id in range(trial_len):
        # extract baseline signals
        trial_baseline_sample = mat_data['DREAMER'][0, 0]['Data'][
            0, subject]['EEG'][0, 0]['baseline'][0, 0][trial_id, 0]
        trial_baseline_sample = trial_baseline_sample[:, :channel_num].swapaxes(
            1, 0)  # channel(14), timestep(61*128)
        trial_baseline_sample = trial_baseline_sample[:, :baseline_num *
                                                      baseline_chunk_size].reshape(
                                                          channel_num,
                                                          baseline_num,
                                                          baseline_chunk_size
                                                      ).mean(
                                                          axis=1
                                                      )  # channel(14), timestep(128)

        # record the common meta info
        trial_meta_info = {'subject_id': subject, 'trial_id': trial_id}

        trial_meta_info['valence'] = mat_data['DREAMER'][0, 0]['Data'][
            0, subject]['ScoreValence'][0, 0][trial_id, 0]
        trial_meta_info['arousal'] = mat_data['DREAMER'][0, 0]['Data'][
            0, subject]['ScoreArousal'][0, 0][trial_id, 0]
        trial_meta_info['dominance'] = mat_data['DREAMER'][0, 0]['Data'][
            0, subject]['ScoreDominance'][0, 0][trial_id, 0]

        # extract experimental signals
        start_at = 0
        end_at = chunk_size

        trial_samples = mat_data['DREAMER'][0, 0]['Data'][0, subject]['EEG'][
            0, 0]['stimuli'][0, 0][trial_id, 0]
        trial_samples = trial_samples[:, :channel_num].swapaxes(
            1, 0)  # channel(14), timestep(n*128)

        while end_at <= trial_samples.shape[1]:
            clip_sample = trial_samples[:, start_at:end_at]

            t_eeg = clip_sample
            t_baseline = trial_baseline_sample

            if not transform is None:
                t = transform(eeg=clip_sample, baseline=trial_baseline_sample)
                t_eeg = t['eeg']
                t_baseline = t['baseline']

            # put baseline signal into IO
            if not 'baseline_id' in trial_meta_info:
                trial_base_id = f'{subject}_{write_pointer}'
                queue.put({'eeg': t_baseline, 'key': trial_base_id})
                write_pointer += 1
                trial_meta_info['baseline_id'] = trial_base_id

            clip_id = f'{subject}_{write_pointer}'

            queue.put({'eeg': t_eeg, 'key': clip_id})
            write_pointer += 1

            # record meta info for each signal
            record_info = {
                'start_at': start_at,
                'end_at': end_at,
                'clip_id': clip_id
            }
            record_info.update(trial_meta_info)
            write_info_fn(record_info)

            start_at = start_at + step
            end_at = start_at + chunk_size


def io_consumer(write_eeg_fn, queue):
    while True:
        item = queue.get()
        if not item is None:
            eeg = item['eeg']
            key = item['key']
            write_eeg_fn(eeg, key)
        else:
            break


class SingleProcessingQueue:
    def __init__(self, write_eeg_fn):
        self.write_eeg_fn = write_eeg_fn

    def put(self, item):
        eeg = item['eeg']
        key = item['key']
        self.write_eeg_fn(eeg, key)


def dreamer_constructor(mat_path: str = './DREAMER.mat',
                        chunk_size: int = 128,
                        overlap: int = 0,
                        channel_num: int = 14,
                        baseline_num: int = 61,
                        baseline_chunk_size: int = 128,
                        transform: Union[None, Callable] = None,
                        io_path: str = './io/dreamer',
                        num_worker: int = 0,
                        verbose: bool = True,
                        cache_size: int = 64 * 1024 * 1024 * 1024) -> None:
    # init IO
    meta_info_io_path = os.path.join(io_path, 'info.csv')
    eeg_signal_io_path = os.path.join(io_path, 'eeg')

    if os.path.exists(io_path) and not os.path.getsize(meta_info_io_path) == 0:
        print(
            f'The target folder already exists, if you need to regenerate the database IO, please delete the path {io_path}.'
        )
        return

    os.makedirs(io_path, exist_ok=True)

    info_io = MetaInfoIO(meta_info_io_path)
    eeg_io = EEGSignalIO(eeg_signal_io_path, cache_size=cache_size)

    # access the dataset files
    mat_data = scio.loadmat(mat_path, verify_compressed_data_integrity=False)

    subject_len = len(mat_data['DREAMER'][0, 0]['Data'][0])  # 23
    trial_len = len(
        mat_data['DREAMER'][0, 0]['Data'][0, 0]['EEG'][0,
                                                       0]['stimuli'][0,
                                                                     0])  # 18

    if verbose:
        # show process bar
        pbar = tqdm(total=subject_len)
        pbar.set_description("[DREAMER]")

    if num_worker > 1:
        manager = Manager()
        queue = manager.Queue(maxsize=MAX_QUEUE_SIZE)
        io_consumer_process = Process(target=io_consumer,
                                      args=(eeg_io.write_eeg, queue),
                                      daemon=True)
        io_consumer_process.start()

        partial_mp_fn = partial(transform_producer,
                                trial_len=trial_len,
                                mat_data=mat_data,
                                chunk_size=chunk_size,
                                overlap=overlap,
                                channel_num=channel_num,
                                baseline_num=baseline_num,
                                baseline_chunk_size=baseline_chunk_size,
                                transform=transform,
                                write_info_fn=info_io.write_info,
                                queue=queue)

        for _ in Pool(num_worker).imap(partial_mp_fn, list(range(subject_len))):
            if verbose:
                pbar.update(1)

        queue.put(None)

        io_consumer_process.join()
        io_consumer_process.close()

    else:
        for subject in list(range(subject_len)):
            transform_producer(subject=subject,
                               trial_len=trial_len,
                               mat_data=mat_data,
                               chunk_size=chunk_size,
                               overlap=overlap,
                               channel_num=channel_num,
                               baseline_num=baseline_num,
                               baseline_chunk_size=baseline_chunk_size,
                               transform=transform,
                               write_info_fn=info_io.write_info,
                               queue=SingleProcessingQueue(eeg_io.write_eeg))
            if verbose:
                pbar.update(1)

    if verbose:
        pbar.close()
        print('Please wait for the writing process to complete...')



from typing import Callable, Dict, Tuple, Union

from ...constants.emotion_recognition.bci2022 import (
    BCI2022_ADJACENCY_MATRIX, BCI2022_CHANNEL_LOCATION_DICT)
from ...functional.emotion_recognition.bci2022 import bci2022_constructor
from ..base_dataset import BaseDataset


class BCI2022Dataset(BaseDataset):
    r'''
    The 2022 EMOTION_BCI competition aims at tackling the cross-subject emotion recognition challenge and provides participants with a batch of EEG data from 80 participants with known emotional state information. Participants are required to establish an EEG computing model with cross-individual emotion recognition ability. The subjects' EEG data were used for real-time emotion recognition. This class generates training samples and test samples according to the given parameters and caches the generated results in a unified input and output format (IO). The relevant information of the dataset is as follows:

    - Author: Please refer to the downloaded URL.
    - Year: 2022
    - Download URL: https://oneuro.cn/n/competitiondetail/2022_emotion_bci/doc0
    - Reference: Please refer to the downloaded URL.
    - Stimulus: video clips.
    - Signals: Electroencephalogram (30 channels at 250Hz) and two channels of left/right mastoid signals from 80 subjects.
    - Rating: 28 video clips are annotated in valence and discrete emotion dimensions. The valence is divided into positive (-1), negative (0), and neutral (1). Discrete emotions are divided into anger (0), disgust (1), fear (2), sadness (3), neutral (4), amusement (5), excitation (6), happiness (7), and warmth (8).

    In order to use this dataset, the download folder :obj:`TrainSet` is required, containing the following files:
    
    - TrainSet_first_batch

        + sub1
        + sub10
        + sub11
        + ...

    - TrainSet_second_batch

        + sub55
        + sub57
        + sub59
        + ...

    An example dataset for CNN-based methods:

    .. code-block:: python

        dataset = BCI2022Dataset(io_path=f'./bci2022',
                              root_path='./TrainSet',
                              offline_transform=transforms.Compose([
                                  transforms.BandDifferentialEntropy(),
                                  transforms.ToGrid(BCI2022_CHANNEL_LOCATION_DICT)
                              ]),
                              online_transform=transforms.ToTensor(),
                              label_transform=transforms.Select(['emotion']))
        print(dataset[0])
        # EEG signal (torch.Tensor[250, 8, 9]),
        # coresponding baseline signal (torch.Tensor[250, 8, 9]),
        # label (int)

    Another example dataset for CNN-based methods:

    .. code-block:: python

        dataset = BCI2022Dataset(io_path=f'./bci2022',
                              root_path='./TrainSet',
                              online_transform=transforms.Compose([
                                  transforms.ToTensor(),
                                  transforms.To2d()
                              ]),
                              label_transform=transforms.Select(['emotion']))
        print(dataset[0])
        # EEG signal (torch.Tensor[30, 250]),
        # coresponding baseline signal (torch.Tensor[30, 250]),
        # label (int)

    An example dataset for GNN-based methods:

    .. code-block:: python
    
        dataset = BCI2022Dataset(io_path=f'./bci2022',
                              root_path='./TrainSet',
                              online_transform=transforms.Compose([
                                  transforms.ToG(BCI2022_ADJACENCY_MATRIX)
                              ]),
                              label_transform=transforms.Select(['emotion']))
        print(dataset[0])
        # EEG signal (torch_geometric.data.Data),
        # coresponding baseline signal (torch_geometric.data.Data),
        # label (int)
        
    In particular, TorchEEG utilizes the producer-consumer model to allow multi-process data preprocessing. If your data preprocessing is time consuming, consider increasing :obj:`num_worker` for higher speedup. If running under Windows, please use the proper idiom in the main module:

    .. code-block:: python
    
        if __name__ == '__main__':
            dataset = BCI2022Dataset(io_path=f'./bci2022',
                              root_path='./TrainSet',
                              online_transform=transforms.Compose([
                                  transforms.ToG(BCI2022_ADJACENCY_MATRIX)
                              ]),
                              label_transform=transforms.Select(['emotion']),
                              num_worker=4)
            print(dataset[0])
            # EEG signal (torch_geometric.data.Data),
            # coresponding baseline signal (torch_geometric.data.Data),
            # label (int)

    Args:
        root_path (str): Downloaded data files in pickle (the TrainSet folder in unzipped 2022EmotionPublic.zip) formats (default: :obj:`'./TrainSet'`)
        chunk_size (int): Number of data points included in each EEG chunk as training or test samples. (default: :obj:`250`)
        overlap (int): The number of overlapping data points between different chunks when dividing EEG chunks. (default: :obj:`0`)
        channel_num (int): Number of channels used, of which the first 30 channels are EEG signals. (default: :obj:`30`)
        online_transform (Callable, optional): The transformation of the EEG signals and baseline EEG signals. The input is a :obj:`np.ndarray`, and the ouput is used as the first and second value of each element in the dataset. (default: :obj:`None`)
        offline_transform (Callable, optional): The usage is the same as :obj:`online_transform`, but executed before generating IO intermediate results. (default: :obj:`None`)
        label_transform (Callable, optional): The transformation of the label. The input is an information dictionary, and the ouput is used as the third value of each element in the dataset. (default: :obj:`None`)
        io_path (str): The path to generated unified data IO, cached as an intermediate result. (default: :obj:`./io/bci2022`)
        num_worker (str): How many subprocesses to use for data processing. (default: :obj:`0`)
        verbose (bool): Whether to display logs during processing, such as progress bars, etc. (default: :obj:`True`)
    
    '''
    channel_location_dict = BCI2022_CHANNEL_LOCATION_DICT
    adjacency_matrix = BCI2022_ADJACENCY_MATRIX

    def __init__(self,
                 root_path: str = './2022EmotionPublic/TrainSet/',
                 chunk_size: int = 250,
                 overlap: int = 0,
                 channel_num: int = 30,
                 online_transform: Union[None, Callable] = None,
                 offline_transform: Union[None, Callable] = None,
                 label_transform: Union[None, Callable] = None,
                 io_path: str = './io/bci2022',
                 num_worker: int = 0,
                 verbose: bool = True,
                 cache_size: int = 64 * 1024 * 1024 * 1024):
        bci2022_constructor(root_path=root_path,
                            chunk_size=chunk_size,
                            overlap=overlap,
                            channel_num=channel_num,
                            transform=offline_transform,
                            io_path=io_path,
                            num_worker=num_worker,
                            verbose=verbose,
                            cache_size=cache_size)
        super().__init__(io_path)

        self.root_path = root_path
        self.chunk_size = chunk_size
        self.overlap = overlap
        self.channel_num = channel_num
        self.online_transform = online_transform
        self.offline_transform = offline_transform
        self.label_transform = label_transform
        self.num_worker = num_worker
        self.verbose = verbose
        self.cache_size = cache_size

    def __getitem__(self, index: int) -> Tuple[any, any, int, int, int]:
        info = self.info.iloc[index].to_dict()

        eeg_index = str(info['clip_id'])
        eeg = self.eeg_io.read_eeg(eeg_index)

        signal = eeg
        label = info

        if self.online_transform:
            signal = self.online_transform(eeg=eeg)['eeg']

        if self.label_transform:
            label = self.label_transform(y=info)['y']

        return signal, label

    @property
    def repr_body(self) -> Dict:
        return dict(
            super().repr_body, **{
                'root_path': self.root_path,
                'chunk_size': self.chunk_size,
                'overlap': self.overlap,
                'channel_num': self.channel_num,
                'online_transform': self.online_transform,
                'offline_transform': self.offline_transform,
                'label_transform': self.label_transform,
                'num_worker': self.num_worker,
                'verbose': self.verbose,
                'cache_size': self.cache_size
            })


import os

from typing import Dict

from torch.utils.data import Dataset
from torcheeg.io import EEGSignalIO, MetaInfoIO


class BaseDataset(Dataset):
    channel_location_dict = {}
    adjacency_matrix = []

    def __init__(self, io_path: str):
        if not self.exist(io_path):
            raise RuntimeError('Database IO does not exist, please regenerate database IO.')
        self.io_path = io_path

        meta_info_io_path = os.path.join(self.io_path, 'info.csv')
        eeg_signal_io_path = os.path.join(self.io_path, 'eeg')

        info_io = MetaInfoIO(meta_info_io_path)
        self.eeg_io = EEGSignalIO(eeg_signal_io_path)

        self.info = info_io.read_all()

    def exist(self, io_path: str) -> bool:
        meta_info_io_path = os.path.join(io_path, 'info.csv')
        eeg_signal_io_path = eeg_signal_io_path = os.path.join(io_path, 'eeg')

        return os.path.exists(meta_info_io_path) and os.path.exists(eeg_signal_io_path)

    def __getitem__(self, index: int) -> any:
        raise NotImplementedError("Method __getitem__ is not implemented in class " + self.__class__.__name__)

    def __len__(self):
        return len(self.info)

    def __copy__(self) -> 'BaseDataset':
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        eeg_signal_io_path = os.path.join(self.io_path, 'eeg')

        result.eeg_io = EEGSignalIO(eeg_signal_io_path)
        return result

    @property
    def repr_body(self) -> Dict:
        return {'io_path': self.io_path}

    @property
    def repr_tail(self) -> Dict:
        return {'length': self.__len__()}

    def __repr__(self) -> str:
        # init info
        format_string = self.__class__.__name__ + '('
        for i, (k, v) in enumerate(self.repr_body.items()):
            # line end
            if i:
                format_string += ','
            format_string += '\n'
            # str param
            if isinstance(v, str):
                format_string += f"    {k}='{v}'"
            else:
                format_string += f"    {k}={v}"
        format_string += '\n)'
        # other info
        format_string += '\n'
        for i, (k, v) in enumerate(self.repr_tail.items()):
            if i:
                format_string += ', '
            format_string += f'{k}={v}'
        return format_string


import torch
import torch.nn as nn
import torch.nn.functional as F

from typing import Dict


class FocalLoss(nn.Module):
    r"""
    Focal loss is a loss that adds a factor :math:`(1 - p_t)^{\gamma}` to the standard cross entropy criterion. Setting :math:`\gamma>0` reduces the relative loss for well-classified examples (:math:`p_t > 0.5`), putting more focus on hard, misclassified examples. As experiments demonstrate, the focal loss enables training highly accurate models in the presence of vast numbers of easy negative examples.

    - Paper: Lin T Y, Goyal P, Girshick R, et al. Focal loss for dense object detection[C]//Proceedings of the IEEE international conference on computer vision. 2017: 2980-2988.
    - URL: https://openaccess.thecvf.com/content_ICCV_2017/papers/Lin_Focal_Loss_for_ICCV_2017_paper.pdf
    - Related Project: https://github.com/ailias/Focal-Loss-implement-on-Tensorflow

    .. code-block:: python

        inputs = torch.randn(3, 5, requires_grad=True)
        targets = torch.empty(3, dtype=torch.long).random_(5)
        loss = FocalLoss(num_classes=5)
        output = loss(inputs, targets)

    This version further adds support for multi-classification and label smoothing to the original implementation.

    Args:
        alpha (float): The hyperparameter alpha in the focal loss. (defualt: :obj:`1.0`)
        gamma (float): The hyperparameter gamma in the focal loss. (defualt: :obj:`2.0`)
        reduction (str): Specifies the reduction to apply to the output. Options include none, mean and sum. 'none': no reduction will be applied, 'mean': the sum of the output will be divided by the number of elements in the output, 'sum': the output will be summed.
        num_classes (int): The number of classes to predict. (defualt: :obj:`2`)
    """
    def __init__(
        self,
        alpha: int = 1.0,
        gamma: int = 2.0,
        reduction: str = 'mean',
        label_smooth: float = 0.05,
        num_classes: int = 2,
    ):
        super(FocalLoss, self).__init__()
        self.alpha = alpha
        self.gamma = gamma
        self.reduction = reduction
        self.label_smooth = label_smooth
        self.num_classes = num_classes

    def forward(self, inputs: torch.Tensor,
                targets: torch.Tensor) -> torch.Tensor:
        r'''
        Args:
            inputs (torch.Tensor): The predictions from the model.
            targets (torch.Tensor): The ground-truth labels.

        Returns:
            torch.Tensor: the loss value.
        '''

        targets = F.one_hot(targets, num_classes=self.num_classes)

        if self.label_smooth is not None:
            targets = (1 - self.label_smooth
                       ) * targets + self.label_smooth / self.num_classes

        loss = F.binary_cross_entropy_with_logits(inputs,
                                                  targets,
                                                  reduction='none')

        p_t = torch.exp(-loss)
        loss = self.alpha * (1 - p_t)**self.gamma * loss

        if self.reduction == 'mean':
            return torch.mean(loss)
        elif self.reduction == 'sum':
            return torch.sum(loss)
        elif self.reduction == 'none':
            return loss
        else:
            raise NotImplementedError

    @property
    def repr_body(self) -> Dict:
        return {
            'alpha': self.alpha,
            'gamma': self.gamma,
            'reduction': self.reduction,
            'label_smooth': self.label_smooth,
            'num_classes': self.num_classes
        }

    def __repr__(self) -> str:
        # init info
        format_string = self.__class__.__name__ + '('
        for i, (k, v) in enumerate(self.repr_body.items()):
            # line end
            if i:
                format_string += ', '
            # str param
            if isinstance(v, str):
                format_string += f"{k}='{v}'"
            else:
                format_string += f"{k}={v}"
        format_string += ')'
        return format_string


import os
from functools import partial
from multiprocessing import Manager, Pool, Process, Queue, set_start_method
from typing import Callable, List, Union

import joblib
import numpy as np
from torcheeg.io import EEGSignalIO, MetaInfoIO
from tqdm import tqdm

MAX_QUEUE_SIZE = 1024

FIRST_BATCH_CHANNEL = [
    'FP1', 'FP2', 'FZ', 'F3', 'F4', 'F7', 'F8', 'FC1', 'FC2', 'FC5', 'FC6',
    'CZ', 'C3', 'C4', 'T7', 'T8', 'A1', 'A2', 'CP1', 'CP2', 'CP5', 'CP6', 'PZ',
    'P3', 'P4', 'P7', 'P8', 'PO3', 'PO4', 'OZ', 'O1', 'O2'
]

SECOND_BATCH_CHANNEL = [
    'FP1', 'FP2', 'FZ', 'F3', 'F4', 'F7', 'F8', 'FC1', 'FC2', 'FC5', 'FC6',
    'CZ', 'C3', 'C4', 'T7', 'T8', 'CP1', 'CP2', 'CP5', 'CP6', 'PZ', 'P3', 'P4',
    'P7', 'P8', 'PO3', 'PO4', 'OZ', 'O1', 'O2', 'A2', 'A1'
]

FIRST_TO_SECOND_ORDER = [
    FIRST_BATCH_CHANNEL.index(c) for c in SECOND_BATCH_CHANNEL
]

VALENCE_DICT = {
    1: -1,  # negative
    2: -1,
    3: -1,
    4: -1,
    5: -1,
    6: -1,
    7: -1,
    8: -1,
    9: -1,
    10: -1,
    11: -1,
    12: -1,
    13: 0,  # neutral
    14: 0,
    15: 0,
    16: 0,
    17: 1,  # positive
    18: 1,
    19: 1,
    20: 1,
    21: 1,
    22: 1,
    23: 1,
    24: 1,
    25: 1,
    26: 1,
    27: 1,
    28: 1
}

EMOTION_DICT = {
    1: 0,  # anger	
    2: 0,
    3: 0,
    4: 1,  # disgust
    5: 1,
    6: 1,
    7: 2,  # fear
    8: 2,
    9: 2,
    10: 3,  # sadness
    11: 3,
    12: 3,
    13: 4,  # neutral
    14: 4,
    15: 4,
    16: 4,
    17: 5,  # amusement
    18: 5,
    19: 5,
    20: 6,  # excitation
    21: 6,
    22: 6,
    23: 7,  # happy
    24: 7,
    25: 7,
    26: 8,  # warmth
    27: 8,
    28: 8
}


def transform_producer(file_name: str, root_path: str, chunk_size: int,
                       overlap: int, channel_num: int,
                       transform: Union[List[Callable], Callable, None],
                       reorder: bool, write_info_fn: Callable, queue: Queue):

    subject = file_name  # subject (54)
    samples = joblib.load(
        os.path.join(root_path, file_name,
                     f'{file_name}.pkl'))  # channel(33), timestep(n*250)

    events = samples[-1]
    if reorder:
        samples = samples.take(FIRST_TO_SECOND_ORDER, axis=0)

    # initial pointers
    trial_id = 0
    write_pointer = 0

    video_id = None
    start_at = None
    end_at = None

    # loop for each trial
    for i, event in enumerate(events):

        if event in list(range(1, 29)):
            # Video events 1-28: Different events correspond to different experimental video materials
            video_id = event
        elif event == 240:
            # Current trial video start event 240: This event appears 0.1s after the video event, indicating that the video starts to play.
            start_at = i
        elif event == 241:
            # Current trial video end event 241: This event indicates that the video ends playing. Block end event 243: This event indicates the end of the block.
            end_at = i
            assert (not video_id is None) and (
                not start_at is None
            ), f'Parse event fail for trial {trial_id} with video_id={video_id}, start_at={start_at}, end_at={end_at}!'

            trial_meta_info = {
                'trial_id': trial_id,
                'video_id': video_id,
                'subject_id': subject,
                'valence': VALENCE_DICT[video_id],
                'emotion': EMOTION_DICT[video_id],
            }

            step = chunk_size - overlap
            cur_start_at = start_at
            cur_end_at = start_at + chunk_size

            while cur_end_at <= end_at:
                t_eeg = samples[:channel_num, cur_start_at:cur_end_at]
                if not transform is None:
                    t_eeg = transform(eeg=t_eeg)['eeg']

                clip_id = f'{subject}_{write_pointer}'
                queue.put({'eeg': t_eeg, 'key': clip_id})
                write_pointer += 1

                record_info = {
                    'start_at': cur_start_at,
                    'end_at': cur_end_at,
                    'clip_id': clip_id
                }
                record_info.update(trial_meta_info)
                write_info_fn(record_info)

                cur_start_at = cur_start_at + step
                cur_end_at = cur_start_at + chunk_size

            # prepare for the next trial
            trial_id += 1
            video_id = None
            start_at = None
            end_at = None


def io_consumer(write_eeg_fn, queue):
    while True:
        item = queue.get()
        if not item is None:
            eeg = item['eeg']
            key = item['key']
            write_eeg_fn(eeg, key)
        else:
            break


class SingleProcessingQueue:
    def __init__(self, write_eeg_fn):
        self.write_eeg_fn = write_eeg_fn

    def put(self, item):
        eeg = item['eeg']
        key = item['key']
        self.write_eeg_fn(eeg, key)


def bci2022_constructor(root_path: str = './2022EmotionPublic/TrainSet/',
                        chunk_size: int = 250,
                        overlap: int = 0,
                        channel_num: int = 30,
                        transform: Union[None, Callable] = None,
                        io_path: str = './io/bci2022',
                        num_worker: int = 0,
                        verbose: bool = True,
                        cache_size: int = 64 * 1024 * 1024 * 1024) -> None:
    # init IO
    meta_info_io_path = os.path.join(io_path, 'info.csv')
    eeg_signal_io_path = os.path.join(io_path, 'eeg')

    if os.path.exists(io_path) and not os.path.getsize(meta_info_io_path) == 0:
        print(
            f'The target folder already exists, if you need to regenerate the database IO, please delete the path {io_path}.'
        )
        return
    os.makedirs(io_path, exist_ok=True)

    meta_info_io_path = os.path.join(io_path, 'info.csv')
    eeg_signal_io_path = os.path.join(io_path, 'eeg')

    info_io = MetaInfoIO(meta_info_io_path)
    eeg_io = EEGSignalIO(eeg_signal_io_path, cache_size=cache_size)

    for train_set_batch_idx, train_set_batch in enumerate(
        ['TrainSet_first_batch', 'TrainSet_second_batch']):
        # loop to access the first batch files
        file_list = os.listdir(os.path.join(root_path, train_set_batch))

        if verbose:
            # show process bar
            pbar = tqdm(total=len(file_list))
            pbar.set_description(f"[BCI2022 BATCH {train_set_batch_idx+1}/2]")

        if num_worker > 1:
            manager = Manager()
            queue = manager.Queue(maxsize=MAX_QUEUE_SIZE)
            io_consumer_process = Process(target=io_consumer,
                                          args=(eeg_io.write_eeg, queue),
                                          daemon=True)
            io_consumer_process.start()

            partial_mp_fn = partial(
                transform_producer,
                root_path=os.path.join(root_path, train_set_batch),
                chunk_size=chunk_size,
                overlap=overlap,
                channel_num=channel_num,
                transform=transform,
                reorder=train_set_batch == 'TrainSet_first_batch',
                write_info_fn=info_io.write_info,
                queue=queue)

            for _ in Pool(num_worker).imap(partial_mp_fn, file_list):
                if verbose:
                    pbar.update(1)

            queue.put(None)

            io_consumer_process.join()
            io_consumer_process.close()

        else:
            for file_name in file_list:
                transform_producer(
                    file_name=file_name,
                    root_path=os.path.join(root_path, train_set_batch),
                    chunk_size=chunk_size,
                    overlap=overlap,
                    channel_num=channel_num,
                    transform=transform,
                    reorder=train_set_batch == 'TrainSet_first_batch',
                    write_info_fn=info_io.write_info,
                    queue=SingleProcessingQueue(eeg_io.write_eeg))
                if verbose:
                    pbar.update(1)

        if verbose:
            pbar.close()
            print('Please wait for the writing process to complete...')
